﻿/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;
using System.Diagnostics;
using System.Reflection;


namespace PaintableSimulator.Engine.PaintableComputers
{
	/// <summary>
	/// 
	/// </summary>
	public partial class GenericPaintableComputer : PaintableComputerBase
	{
		#region AbstractPaintableComputer Members

		/// <summary>
		/// Cycles this instance.
		/// </summary>
		public override void Cycle()
		{
			// increase internal clock
			internalClock++;


			// process arrived packets
			while (receivePadQueue.Count != 0)
			{
				// get packet
				NetworkPacket receivedPacket = receivePadQueue.Dequeue();

				switch (receivedPacket.PacketType)
				{
					case PacketType.Homepage:
						{
							bool notFound = true;

							Homepage received = ((HomepagePacket)receivedPacket).Homepage;

							// prebaci homepage u listu mirrora, ako je vec tu, samo zamijeni, inace dodaj u listu
							for (int i = 0; i < mirroredHomepages.Count; i++)
							{
								ExtendedHomepageData ehp = mirroredHomepages[i];

								if (ehp.OriginComputerId == receivedPacket.OriginId)
								{
									ehp = null;
									ehp = new ExtendedHomepageData(receivedPacket.OriginId, received, internalClock);
									notFound = false;


									// When the HP is found exit the for loop
									break;
								}
							}

							// we don't have this mirror yet, so add it! But check if we have enough space first
							if (notFound && (mirroredHomepages.Count < maximumNeighbourCount))
								mirroredHomepages.Add(new ExtendedHomepageData(receivedPacket.OriginId, received, internalClock));


						}
						break;

					case PacketType.ProcessFragment:
						{
							// check if the packet was meant for us
							if (receivedPacket.DestinationId == this.ComputerId)
							{
								ProcessFragmentPacket pfPacket = (ProcessFragmentPacket)receivedPacket;
								ExtendedProcessFragmentData pf = new ExtendedProcessFragmentData(this.maximumNeighbourCount);

								// store the origin for later use
								pf.OriginComputerId = pfPacket.OriginId;

								// create an instance of the ProcessFragment
								pf.ProcessFragmentType = pfPacket.ProcessFragment.GetType();

								pf.ProcessFragment = pfPacket.ProcessFragment;

								// attach it to this computer
								pf.ProcessFragment.Api = (IApiInterface)this;

								// create personal homepage in advance (SPEEDUP HACK)
								if (!localHomepage.ContainsKey(pf.ProcessFragment.ProcessFragmentId))
									localHomepage.Add(pf.ProcessFragment.ProcessFragmentId, new HomepagePostCollection());

								try
								{
#if TRACE
									if (log.IsDebugEnabled)
										log.Debug(String.Format("{0:X}: CALLING INSTALL FOR {1:X}", this.computerId, pf.ProcessFragment.ProcessFragmentId));
#endif
									// call its install method
									currentProcessFragment = pf;
									currentProcessFragment.ProcessFragment.Install();
									currentProcessFragment = null;

									// if install succedes add it to the list of installed process fragments
									installedProcessFragments.Add(pf);
								}
								catch (System.Security.SecurityException sx)
								{
									if (log.IsWarnEnabled)
										log.Warn("Security exception", sx);

									// THIS IS A BAD Process Fragment
									RemoveAllEntries(pf.ProcessFragment.ProcessFragmentId);
								}
								catch
								{
									// Process fragment install method failure

									// clean up his posts
									RemoveAllEntries(pf.ProcessFragment.ProcessFragmentId);
									
									// ignore the fragment
								}

							}

						}
						break;

					default:
						{
						}
						break;
				}

			}



			// process all installed fragments Update method
			for (int i = installedProcessFragments.Count-1; i >= 0; i--)
			{
				currentProcessFragment = installedProcessFragments[i];

				if (currentProcessFragment == null)
					continue;

				// skip processing if the process fragment queued itself for uninstall
				if (currentProcessFragment.QueuedForUninstall)
					continue;

#if TRACE
				if (log.IsDebugEnabled)
					log.Debug(String.Format("{0:X}: CALLING UPDATE FOR {1:X}", this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId));
#endif

				// call update
				try
				{
					currentProcessFragment.ProcessFragment.Update(0);
				}
				catch (System.Security.SecurityException sx)
				{
					if (log.IsWarnEnabled)
						log.Warn("Security exception", sx);

					// THIS IS A BAD Process Fragment
					RemoveAllEntries(currentProcessFragment.ProcessFragment.ProcessFragmentId);
					
					// ToDo: delete pfrag
					installedProcessFragments.RemoveAt(i);
					continue;

				}
				catch (Exception ex)
				{
					Debug.WriteLine(ex.ToString());

					// increase update method failure count
					currentProcessFragment.SubsequentUpdateFails++;

					// if there were more than two susequent update method failures
					// queue this process fragment for uninstall
					if (currentProcessFragment.SubsequentUpdateFails > 2)
					{
						uninstallQueue.Push(byte.MaxValue, currentProcessFragment);
						continue;
					}

				}

				currentProcessFragment.SubsequentUpdateFails = 0;
				currentProcessFragment = null;
			}


			// ProcessFragment transfer requests processing
			while (preTransferQueue.Count != 0)
			{
				TransferRequestTicket request = preTransferQueue.Dequeue();

				// find out if the pfrag has been BAD!
				if (installedProcessFragments.Contains(request.Requestee))
					currentProcessFragment = request.Requestee;
				else
					continue;

				// skip processing if the process fragment queued itself for uninstall
				if (currentProcessFragment.QueuedForUninstall)
					continue;

				// tu treba pogledati da li se dozvoljava transfer prema van ili ne
				if (sendPadQueue.Count > maximumNeighbourCount)
				{
					try
					{
#if TRACE
						if (log.IsDebugEnabled)
							log.Debug(String.Format("{0:X}: CALLING TRANSFERREFUSED FOR {1:X}", this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId));
#endif
						currentProcessFragment.ProcessFragment.TransferRefused();
					}
					catch (System.Security.SecurityException sx)
					{

						if (log.IsWarnEnabled)
							log.Warn("Security exception", sx);

						// THIS IS A BAD Process Fragment
						RemoveAllEntries(currentProcessFragment.ProcessFragment.ProcessFragmentId);
						// ToDo: delete pfrag
						installedProcessFragments.Remove(currentProcessFragment);
						
					}
					catch (Exception ex)
					{
						Debug.WriteLine(ex.ToString());
					}

					// skip the rest of the processing
					continue;
				}


				// Create a copy of the process fragment
				ProcessFragmentBase copy = (ProcessFragmentBase)Activator.CreateInstance(currentProcessFragment.ProcessFragmentType, false);


				try
				{
#if TRACE
					if (log.IsDebugEnabled)
						log.Debug(String.Format("{0:X}: CALLING TRANSFERGRANTED FOR {1:X}", this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId));
#endif

					// call TransferGranted to receive an initial state vector for the transfered pfrag
					currentProcessFragment.ProcessFragment.TransferGranted(ref copy);
				}
				catch (System.Security.SecurityException sx)
				{
					
					if (log.IsWarnEnabled)
						log.Warn("Security exception", sx);

					// THIS IS A BAD Process Fragment
					RemoveAllEntries(currentProcessFragment.ProcessFragment.ProcessFragmentId);
					// ToDo: delete pfrag
					installedProcessFragments.Remove(currentProcessFragment);

					// skip to next request ticket
					continue;

				}
				catch (Exception ex)
				{
					Debug.WriteLine(ex.ToString());
				}

				// check if the pfrag deleted our copy
				if (copy == null)
					continue;

				// create a packet
				ProcessFragmentPacket pfPacket = new ProcessFragmentPacket(
								copy,
								this.computerId,
								request.DestinationId);

				// add the packet to the send queue
				sendPadQueue.Push(request.Priority, pfPacket);

				// cleanup flags and transfer data
				currentProcessFragment.ResetQueuedForTransferFlag(request.MirrorIndex);

			}



			// ProcessFragment uninstall processing
			while (uninstallQueue.Count != 0)
			{
				currentProcessFragment = uninstallQueue.Pop();

				try
				{
#if TRACE
					if (log.IsDebugEnabled)
						log.Debug(String.Format("{0:X}: CALLING UNINSTALL FOR {1:X}", this.computerId, currentProcessFragment.ProcessFragment.ProcessFragmentId));
#endif

					currentProcessFragment.ProcessFragment.Uninstall(0);

				}
				catch (System.Security.SecurityException sx)
				{

					if (log.IsWarnEnabled)
						log.Warn("Security exception", sx);

					// THIS IS A BAD Process Fragment
					RemoveAllEntries(currentProcessFragment.ProcessFragment.ProcessFragmentId);
					// ToDo: delete pfrag
					installedProcessFragments.Remove(currentProcessFragment);
					
					// skip to next uninstall request
					continue;
				}
				catch (Exception ex)
				{
					Debug.WriteLine(ex.ToString());
				}

				installedProcessFragments.Remove(currentProcessFragment);

			}


			// Perriodically update the neighbourhood
			// could be used like a "I'm Alive" packet to control the mirror timeout (not implemented yet)
			internalClock++;
			if (internalClock % homePageResendPeriod == 0)
				localHomepage.HasChanged = true;


			// if the home page changed (either by modification or it's a periodic update)
			// queue a HomepagePacket for broadcast
			if (localHomepage.HasChanged)
			{
				//BROADCAST a HomepagePacket
				HomepagePacket hpPacket = new HomepagePacket(localHomepage.Clone(), computerId, 0);
				sendPadQueue.Push(byte.MaxValue, hpPacket);

				// reset flag
				localHomepage.HasChanged = false;
			}



			// process mirrored homepage timeouts
			for (int i = mirroredHomepages.Count - 1; i >= 0; i--)
				if (internalClock - mirroredHomepages[i].ArrivedOnTick > 100 * homePageResendPeriod)
					mirroredHomepages.RemoveAt(i);
		}


		#endregion
	}
}
