﻿/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System.Collections.Generic;
using System.Diagnostics;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;
using System.Xml;
using System.Globalization;
using System;
using System.Reflection;

namespace PaintableSimulator.Engine.PaintableComputers
{
	public partial class GenericPaintableComputer
	{

		#region TransferRequest Class

		/// <summary>
		/// 
		/// </summary>
		[DebuggerStepThrough()]
		private class TransferRequestTicket
		{
			#region Private Fields

			private int mirrorIndex;
			private uint destinationId;
			private byte priority;
			private ExtendedProcessFragmentData requestee;

			#endregion

			#region Constructors

			/// <summary>
			/// Initializes a new instance of the <see cref="T:TransferRequest"/> class.
			/// </summary>
			/// <param name="priority">The priority.</param>
			/// <param name="mirrorIndex">Index of the mirror.</param>
			/// <param name="destinationId">The destination id.</param>
			/// <param name="requestee">The requestee.</param>
			public TransferRequestTicket(byte priority, int mirrorIndex, uint destinationId,
										 ExtendedProcessFragmentData requestee)
			{
				this.mirrorIndex = mirrorIndex;
				this.priority = priority;
				this.destinationId = destinationId;
				this.requestee = requestee;
			}

			#endregion

			#region Properties

			/// <summary>
			/// Gets or sets the index of the mirror.
			/// </summary>
			/// <value>The index of the mirror.</value>
			public int MirrorIndex
			{
				get { return mirrorIndex; }
				set { mirrorIndex = value; }
			}

			/// <summary>
			/// Gets or sets the destination id.
			/// </summary>
			/// <value>The destination id.</value>
			public uint DestinationId
			{
				get { return destinationId; }
				set { destinationId = value; }
			}

			/// <summary>
			/// Gets or sets the requestee.
			/// </summary>
			/// <value>The requestee.</value>
			public ExtendedProcessFragmentData Requestee
			{
				get { return requestee; }
				set { requestee = value; }
			}

			/// <summary>
			/// Gets or sets the priority.
			/// </summary>
			/// <value>The priority.</value>
			public byte Priority
			{
				get { return priority; }
				set { priority = value; }
			}

			#endregion
		}

		#endregion

		#region Private Fields

		/// <summary>
		/// 
		/// </summary>
		private Queue<TransferRequestTicket> preTransferQueue;


		/// <summary>
		/// 
		/// </summary>
		private List<ExtendedProcessFragmentData> installedProcessFragments;

		/// <summary>
		/// 
		/// </summary>
		protected uint internalClock;

		/// <summary>
		/// 
		/// </summary>
		private const int homePageResendPeriod = 17;

		/// <summary>
		/// Currently executing process fragment.
		/// </summary>
		private ExtendedProcessFragmentData currentProcessFragment;

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="T:GenericPaintableComputer"/> class.
		/// </summary>
		public GenericPaintableComputer()
			: this(DEFAULT_MAXIMUMNEIGHBOURCOUNT)
		{
		}


		/// <summary>
		/// Initializes a new instance of the <see cref="T:GenericPaintableComputer"/> class.
		/// </summary>
		/// <param name="maximumNeighbourCount">The maximum neighbour count.</param>
		public GenericPaintableComputer(uint maximumNeighbourCount)
			: base(maximumNeighbourCount)
		{
			preTransferQueue = new Queue<TransferRequestTicket>();
			installedProcessFragments = new List<ExtendedProcessFragmentData>();
		}

		#endregion

	}
}