﻿/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System.Diagnostics;
using System;
using System.Reflection;

namespace PaintableSimulator.Engine.PaintableComputers.OperatingSystem
{
	/// <summary>
	/// A packet type enumeration
	/// </summary>
	public enum PacketType
	{
		/// <summary>
		/// Designates a packet to carry ProcessFragment information
		/// </summary>
		ProcessFragment,

		/// <summary>
		/// Designates a packet to carry HomePage information
		/// </summary>
		Homepage
	}


	/// <summary>
	/// 
	/// </summary>
	[DebuggerStepThrough()]
	public abstract class NetworkPacket
	{
		private PacketType packetType;

		/// <summary>
		/// Gets or sets the type of the packet.
		/// </summary>
		/// <value>The type of the packet.</value>
		public PacketType PacketType
		{
			get { return packetType; }
		}


		private uint originId;

		/// <summary>
		/// Gets or sets the origin id.
		/// </summary>
		/// <value>The origin id.</value>
		public uint OriginId
		{
			get { return originId; }
			set { originId = value; }
		}

		private uint destinationId;

		/// <summary>
		/// Gets or sets the destination id.
		/// </summary>
		/// <value>The destination id.</value>
		public uint DestinationId
		{
			get { return destinationId; }
			set { destinationId = value; }
		}


		/// <summary>
		/// Initializes a new instance of the <see cref="NetworkPacket"/> class.
		/// </summary>
		/// <param name="type">The type.</param>
		/// <param name="originId">The origin id.</param>
		/// <param name="destinationId">The destination id.</param>
		protected NetworkPacket(PacketType type, uint originId, uint destinationId)
		{
			packetType = type;
			this.originId = originId;
			this.destinationId = destinationId;
		}

	}


	/// <summary>
	/// 
	/// </summary>
	[DebuggerStepThrough()]
	public sealed class ProcessFragmentPacket : NetworkPacket
	{
		private ProcessFragmentBase processFragment;


		/// <summary>
		/// Gets or sets the process fragment.
		/// </summary>
		/// <value>The process fragment.</value>
		public ProcessFragmentBase ProcessFragment
		{
			get { return processFragment; }
			set { processFragment = value; }
		}


		/// <summary>
		/// Initializes a new instance of the <see cref="ProcessFragmentPacket"/> class.
		/// </summary>
		/// <param name="processFragment">The process fragment.</param>
		/// <param name="originId">The origin id.</param>
		/// <param name="destinationId">The destination id.</param>
		public ProcessFragmentPacket(ProcessFragmentBase processFragment, uint originId, uint destinationId)
			: base(PacketType.ProcessFragment, originId, destinationId)
		{
			this.processFragment = processFragment;
		}


	}

	/// <summary>
	/// 
	/// </summary>
	[DebuggerStepThrough()]
	public sealed class HomepagePacket : NetworkPacket
	{
		private Homepage homepage;

		/// <summary>
		/// Gets or sets the home page.
		/// </summary>
		/// <value>The home page.</value>
		public Homepage Homepage
		{
			get { return homepage; }
			set { homepage = value; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="HomepagePacket"/> class.
		/// </summary>
		/// <param name="homepage">The homepage.</param>
		/// <param name="originId">The origin id.</param>
		/// <param name="destinationId">The destination id.</param>
		public HomepagePacket(Homepage homepage, uint originId, uint destinationId)
			: base(PacketType.Homepage, originId, destinationId)
		{
			this.homepage = homepage;
		}

	}
}