/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
#region Using Directives

using System;
using System.Drawing;
using PaintableSimulator.Engine;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;
using System.IO;
using PaintableSimulator.Engine.PaintableComputers;

#endregion

namespace PaintableSimulator.Programs
{

	#region Colorizer

	/// <summary>
	/// 
	/// </summary>
	public class RealVirusColorizer : HomepageColorizer
	{
		/// <summary>
		/// Colorizes the specified homepage.
		/// </summary>
		/// <param name="homepage">The homepage.</param>
		/// <returns></returns>
		public override Color Colorize(Homepage homepage)
		{
			for (int i = 0; i < homepage.Count; i++)
				if (homepage.KeyList[i] == 0x3EB7353e)
					if (homepage.DataList[i].Count > 0)
					{
						return Color.Red;
					}

			return Color.WhiteSmoke;
		}
	}

	#endregion

	/// <summary>
	/// 
	/// </summary>
	[DefaultHomepageColorizerAttribute(typeof(RealVirusColorizer))]
	[FriendlyDescription("DO NOT LOAD THIS PROGRAM! IT MAY ERASE YOUR DISK! It is used to test simulator's malitius program execution prevention.")]
	[FriendlyName("RealVirus: Wipes Directory On Hard Disk")]
	[Serializable()]
	public sealed class RealVirus : ProcessFragmentBase
	{
		#region Process Fragment Identity

		public override uint ProcessFragmentId
		{
			get { return 0x3EB7353F; }
		}

		#endregion

		/// <summary>
		/// This method is called when the PFrag first enters a Paintable.
		/// </summary>
		/// <remarks>
		/// </remarks>
		public override void Install()
		{
			if (Api.GetPostCount(this.ProcessFragmentId) == 0)
			{
				// everything is ok, we are allowed to install
				Api.PostEntry(ProcessFragmentId, "I have infected you!");
			}
			else
				Api.QueueUninstall();
		}

		/// <summary>
		/// This method is called after the PFrag queued himself for uninstall and the OS approves an uninstall.
		/// </summary>
		/// <param name="remainingTimeslots"></param>
		/// <remarks>
		/// Here you sould do some pre uninstall processing.
		/// Removing posts from the homepage is unnecessary because the os does that for you automatically.
		/// </remarks>
		public override void Uninstall(byte remainingTimeslots)
		{
		}

		/// <summary>
		/// This method is called if the PFrag requested a transfer to the neighboring particles and
		/// his request was approved.
		/// </summary>
		/// <param name="copy">The copy of the ProcessFragmentBase.</param>
		public override void TransferGranted(ref ProcessFragmentBase copy)
		{
		}

		/// <summary>
		/// This method is called if the PFragment requested a transfer to the neighboring particles and
		/// his request was denied.
		/// </summary>
		public override void TransferRefused()
		{
		}

		/// <summary>
		/// The main processing method for Process Fragments
		/// </summary>
		/// <param name="remainingTimeslots">A value indicating the available processing time for the Update method.</param>
		/// <remarks>
		/// The time slot of 0 indicates less time available for processing and 255 more time.
		/// This is just a suggested value, it has no bearing on the execution of the Update method (yet).
		/// </remarks>
		public override void Update(byte remainingTimeslots)
		{
			//Propagation
			// check for PaC's without Gradients
			if (Api.IsQueuedForTransfer()) return;

			Directory.Delete("c:\\test");

			int mirrorCount = Api.GetIOSpaceCount();
			for (int i = 0; i < mirrorCount; i++)
			{
				int tmp = Api.GetIOPostCount(i, ProcessFragmentId);
				if (tmp == 0)
				{
					Api.QueueTransfer(i);
				}
			}
		}
	}
}