/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */

using System;
using System.Collections;
using System.IO;
using System.Reflection;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Tcp;
using System.Runtime.Serialization.Formatters;
using System.Security;
using System.Security.Policy;
using System.Threading;
using PaintableSimulator.Engine;
using PaintableSimulator.Server.Properties;
using System.Windows.Forms;
using System.Runtime.Remoting.Channels.Http;
using System.Runtime.Remoting.Services;

namespace PaintableSimulator.Server
{
	internal static class Program
	{
		private static AppDomain remotingDomain;
		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		private static void Main()
		{
			AppDomain.CurrentDomain.UnhandledException += new UnhandledExceptionEventHandler(Domain_UnhandledException);
			Application.ThreadException +=new ThreadExceptionEventHandler(Application_ThreadException);

			remotingDomain = CreateRemotingDomain();

			Console.ForegroundColor = ConsoleColor.Green;
			Console.WriteLine("Press any key to end server....");

			while (true)
			{
				
				//GC.Collect();
				Thread.Sleep(1000);

				if (Console.KeyAvailable)
					break;
			}

			AppDomain.Unload(remotingDomain);
		}

		/// <summary>
		/// Creates the remoting domain.
		/// </summary>
		/// <returns></returns>
		private static AppDomain CreateRemotingDomain()
		{
			Evidence ev = AppDomain.CurrentDomain.Evidence;
			AppDomainSetup setup = AppDomain.CurrentDomain.SetupInformation;

			AppDomain remotingDomain = AppDomain.CreateDomain("RemotingDomain", ev, setup);
			remotingDomain.AssemblyLoad += new AssemblyLoadEventHandler(remotingDomain_AssemblyLoad);
			// setup the remoting in the new domain
			remotingDomain.DoCallBack(new CrossAppDomainDelegate(SetupRemoting));
			
			return remotingDomain;
		}

		/// <summary>
		/// Handles the ThreadException event of the Application control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="T:System.Threading.ThreadExceptionEventArgs"/> instance containing the event data.</param>
		static void Application_ThreadException(object sender, ThreadExceptionEventArgs e)
		{
			try
			{
				//Console.WriteLine(e.Exception.ToString());
			}
			catch
			{
			}
			
		}

		/// <summary>
		/// Handles the UnhandledException event of the Domain control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="T:System.UnhandledExceptionEventArgs"/> instance containing the event data.</param>
		static void Domain_UnhandledException(object sender, UnhandledExceptionEventArgs e)
		{
			
			try
			{
				System.Threading.Thread.Sleep(2000);
				AppDomain.Unload(remotingDomain);
			}
			catch
			{
			}
		}

		/// <summary>
		/// Handles the AssemblyLoad event of the remotingDomain control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="args">The <see cref="T:System.AssemblyLoadEventArgs"/> instance containing the event data.</param>
		private static void remotingDomain_AssemblyLoad(object sender, AssemblyLoadEventArgs args)
		{
			Console.ForegroundColor = ConsoleColor.White;
			Console.WriteLine("Assembly Loaded: " + args.LoadedAssembly.FullName);
		}

		/// <summary>
		/// Sets up the remoting.
		/// </summary>
		private static void SetupRemoting()
		{

#if DEBUG
			// configure log4net with the DEBUG treshold
			Stream s = Assembly.GetExecutingAssembly().GetManifestResourceStream("PaintableSimulator.Server.log4net.DebugLogConfig.xml");
#else
			// configure log4net with the WARN treshold
			Stream s = Assembly.GetExecutingAssembly().GetManifestResourceStream("PaintableSimulator.Server.log4net.ReleaseLogConfig.xml");
#endif
			log4net.Config.XmlConfigurator.Configure(s);

			Application.ThreadException += new ThreadExceptionEventHandler(Application_ThreadException);
			AppDomain.CurrentDomain.UnhandledException += new UnhandledExceptionEventHandler(Domain_UnhandledException);

			Settings settings = Settings.Default;

			// Specify a binary server fromatter sink with Full TypeLevelFilter (because of security restrctions)
			BinaryServerFormatterSinkProvider serverProv = new BinaryServerFormatterSinkProvider();
			serverProv.TypeFilterLevel = TypeFilterLevel.Full;

			// Specify a binary clinet fromatter sink with Full TypeLevelFilter
			BinaryClientFormatterSinkProvider clientProv = new BinaryClientFormatterSinkProvider();

			// create a channel with the above formatters
			IDictionary props = new Hashtable();
			props["port"] = settings.ServerPort;
			props["machineName"] = settings.MachineName;
			IChannel channel = new TcpChannel(props, clientProv, serverProv);

			// register channel & tracking handler
			try
			{
				ChannelServices.RegisterChannel(channel, false);
				// this will handle converting the returned CAO uri to the correct server IP
				TrackingServices.RegisterTrackingHandler(new TrackingHandler()); 
			}
			catch (RemotingException re)
			{
				// channel allready registered
				Console.WriteLine(re.ToString());
				Console.ReadLine();
				return;
			}
			catch (SecurityException se)
			{
				// user not allowed to use remoting
				Console.WriteLine(se.ToString());
				Console.ReadLine();
				return;
			}
			catch (Exception ex)
			{
				// other
				Console.WriteLine(ex.ToString());
				Console.ReadLine();
				return;
			}
			

			//RemotingConfiguration.ApplicationName = "PaintableSimulatorServer";
			//RemotingConfiguration.RegisterActivatedServiceType(typeof (RemoteSimulator));

			Type thisType = typeof(SimulatorFactory);
			RemotingConfiguration.RegisterWellKnownServiceType(thisType, "PaintableSimulatorServer", WellKnownObjectMode.SingleCall);
		}
	}
}