/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using PaintableSimulator.Engine;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;
using System.Runtime.Remoting;
using System.Net.Sockets;
using PaintableSimulator.Engine.PaintableComputers;

namespace PaintableSimulator.SmartClient
{

	/// <summary>
	/// 
	/// </summary>
	public partial class DebugVisualizer : Form
	{

		/// <summary>
		/// 
		/// </summary>
		public delegate void RefreshDebugInfoHandler();

		/// <summary>
		/// 
		/// </summary>
		private PaintableComputerBase debugComputer;
		private SimpleComputerInfo debugComputerInfo;

		/// <summary>
		/// 
		/// </summary>
		public int ComputerIndex;

		private List<Homepage> mirrors;
		private int selectedMirrorIndex;

		/// <summary>
		/// 
		/// </summary>
		private Homepage localHomepage;


		private SimulatorClient parent;

		/// <summary>
		/// Initializes a new instance of the <see cref="T:DebugVisualizer"/> class.
		/// </summary>
		/// <param name="computerIndex">Index of the computer in the simulator's list.</param>
		/// <param name="computerInfo">The computer info.</param>
		/// <param name="debugComputer">The debug computer.</param>
		/// <param name="parent">The parent.</param>
		public DebugVisualizer(int computerIndex, SimpleComputerInfo computerInfo, PaintableComputerBase debugComputer, SimulatorClient parent)
		{
			this.ComputerIndex = computerIndex;
			this.debugComputer = debugComputer;
			this.debugComputerInfo = computerInfo;
			this.localHomepage = debugComputer.LocalHomepage;
			this.parent = parent;
			this.mirrors = debugComputer.MirroredHomepages;
			this.selectedMirrorIndex = 0;

			InitializeComponent();

			string info = String.Format(" - ID:{0} - {1}", debugComputer.ComputerId, debugComputerInfo.Location.ToString());
			this.Text += info;
		}


		/// <summary>
		/// Refreshes the debug info.
		/// </summary>
		public void RefreshDebugInfo()
		{

			if (debugComputer == null)
				return;

			// get mirrors
			mirrors = debugComputer.MirroredHomepages;
			if (mirrors.Count != 0)
			{
				// mirror selector
				mirrorSelector.Items.Clear();
				for (int i = 0; i < mirrors.Count; i++)
					mirrorSelector.Items.Add("Mirror " + i.ToString());

				if (mirrorSelector.SelectedIndex < 0)
					mirrorSelector.SelectedIndex = 0;
				if (selectedMirrorIndex > mirrors.Count)
				{
					mirrorSelector.SelectedIndex = 0;
					selectedMirrorIndex = 0;
				}
			}


			// refresh homepages
			RefreshLocalHomepage();
			RefreshMirroredHomepage();

		}

		/// <summary>
		/// Refreshes the mirrored homepage.
		/// </summary>
		private void RefreshMirroredHomepage()
		{
			// prevent mirror logic if no mirrors
			if (mirrors.Count == 0)
				return;


			debugMirrorHomepage.BeginUpdate();
			debugMirrorHomepage.Nodes.Clear();

			Homepage homepage = mirrors[selectedMirrorIndex];

			try
			{

				if (homepage != null)
				{
					string processFragmentLabelTemplate = "ProcessFragment = {0}";
					string postLabelTemplate = "Post[{0}] = {1}";
					string arrayItemLabelTemplate = "Item[{0}] = {1}";

					// loop through process Fragment lists
					for (int i = 0; i < homepage.KeyList.Count; i++)
					{
						string pfragName = "";
						uint key = homepage.KeyList[i];

						if (Program.simulator.Plugins.ProcessFragmentIdTranslations.ContainsKey(key))
							// display name
							pfragName = Program.simulator.Plugins.ProcessFragmentIdTranslations[key];
						else
							// display hex key
							pfragName = homepage.KeyList[i].ToString("X");

						TreeNode fragmentPostsNode = new TreeNode(String.Format(processFragmentLabelTemplate, pfragName));

						// looop through process fragment post list
						for (int j = 0; j < homepage[homepage.KeyList[i]].Count; j++)
						{
							string label;

							if (homepage[homepage.KeyList[i]][j] != null)
								label = String.Format(postLabelTemplate, j, homepage[homepage.KeyList[i]][j].ToString());
							else
								label = String.Format(postLabelTemplate, j, "null");

							TreeNode postNode = new TreeNode(label);

							// loop through process fragment post array list if possible
							if ((homepage[homepage.KeyList[i]][j] != null) && (homepage[homepage.KeyList[i]][j] is Array))
							{
								Array array = (Array)homepage[homepage.KeyList[i]][j];
								if (array.Rank == 1)
									for (int k = 0; k < array.Length; k++)
									{
										string arrayLabel = String.Format(arrayItemLabelTemplate, k, array.GetValue(k).ToString());
										postNode.Nodes.Add(new TreeNode(array.GetValue(k).ToString()));

									}
								else
									postNode.Nodes.Add(new TreeNode("UNABLE TO DISPLAY"));
							}


							fragmentPostsNode.Nodes.Add(postNode);
						}

						debugMirrorHomepage.Nodes.Add(fragmentPostsNode);
					}
				}

			}
			catch (RemotingException ex)
			{
				Program.HandleIrrecoverableRemotingException(ex);
				return;
			}
			catch (SocketException ex)
			{
				Program.HandleIrrecoverableRemotingException(ex);
				return;
			}
			catch (Exception ex)
			{
				if (Program.log.IsErrorEnabled)
					Program.log.Error("Error:", ex);
				return;
			}


			debugMirrorHomepage.ExpandAll();
			debugMirrorHomepage.EndUpdate();

		}



		/// <summary>
		/// Refreshes the local homepage.
		/// </summary>
		private void RefreshLocalHomepage()
		{
			debugHomepageContents.BeginUpdate();
			debugHomepageContents.Nodes.Clear();

			try
			{

				if (localHomepage != null)
				{
					string processFragmentLabelTemplate = "ProcessFragment = {0}";
					string postLabelTemplate = "Post[{0}] = {1}";
					string arrayItemLabelTemplate = "Item[{0}] = {1}";

					// loop through process Fragment lists
					for (int i = 0; i < localHomepage.KeyList.Count; i++)
					{
						string pfragName = "";
						uint key = localHomepage.KeyList[i];

						if (Program.simulator.Plugins.ProcessFragmentIdTranslations.ContainsKey(key))
							// display name
							pfragName = Program.simulator.Plugins.ProcessFragmentIdTranslations[key];
						else
							// display hex key
							pfragName = localHomepage.KeyList[i].ToString("X");

						TreeNode fragmentPostsNode = new TreeNode(String.Format(processFragmentLabelTemplate, pfragName));

						// looop through process fragment post list
						for (int j = 0; j < localHomepage[localHomepage.KeyList[i]].Count; j++)
						{
							string label;

							if (localHomepage[localHomepage.KeyList[i]][j] != null)
								label = String.Format(postLabelTemplate, j, localHomepage[localHomepage.KeyList[i]][j].ToString());
							else
								label = String.Format(postLabelTemplate, j, "null");

							TreeNode postNode = new TreeNode(label);

							// loop through process fragment post array list if possible
							if ((localHomepage[localHomepage.KeyList[i]][j] != null) && (localHomepage[localHomepage.KeyList[i]][j] is Array))
							{
								Array array = (Array)localHomepage[localHomepage.KeyList[i]][j];

								if (array.Rank == 1)
									for (int k = 0; k < array.Length; k++)
									{
										string arrayLabel = String.Format(arrayItemLabelTemplate, k, array.GetValue(k).ToString());
										postNode.Nodes.Add(new TreeNode(array.GetValue(k).ToString()));

									}
								else
									postNode.Nodes.Add("UNABLE TO DISPLAY");
							}


							fragmentPostsNode.Nodes.Add(postNode);
						}

						debugHomepageContents.Nodes.Add(fragmentPostsNode);
					}
				}

			}
			catch (RemotingException ex)
			{
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (SocketException ex)
			{
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (Exception ex)
			{
				System.Diagnostics.Debug.WriteLine(ex.ToString());
				return;
			}


			debugHomepageContents.ExpandAll();
			debugHomepageContents.EndUpdate();
		}


		/// <summary>
		/// Handles the FormClosing event of the DebugVisualizer control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="T:System.Windows.Forms.FormClosingEventArgs"/> instance containing the event data.</param>
		private void DebugVisualizer_FormClosing(object sender, FormClosingEventArgs e)
		{
			parent.RemoveDebugVisualizer(this);
		}

		private void mirrorSelector_SelectedItemChanged(object sender, EventArgs e)
		{
			selectedMirrorIndex = mirrorSelector.SelectedIndex;
		}

	}
}