/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using AviLib;
using PaintableSimulator.Engine;
using Win32;
using HANDLE = System.IntPtr;
using HBRUSH = System.IntPtr;
using HPEN = System.IntPtr;
using HWND = System.IntPtr;
using HDC = System.IntPtr;
using RECT = System.Drawing.Rectangle;
using SIZE = System.Drawing.Size;
using POINT = System.Drawing.Point;

namespace PaintableSimulator.Visualisation
{
	// AVI Export

	#region AVI

	public class AVIExport
	{
		private uint compression = AviCompression.None;
		private string fileName;
		private AviWriter aviFile;

		public bool AviOpened
		{
			get { return (aviFile != null); }
		}


		// Stvaranje AVI Fajla
		public AVIExport()
		{
			aviFile = null;
		}

		~AVIExport()
		{
			CloseAviFile();
		}

		public void CloseAviFile()
		{
			if (aviFile != null)
			{
				aviFile.Close();
			}
			aviFile = null;
		}

		public void CreateAviFile(String FileName, uint Compression)
		{
			// Izbrii staru datoteku
			try
			{
				File.Delete(FileName);
				fileName = FileName;
			}
			finally
			{
			}
			;

			compression = Compression;
		}

		public void SaveFrame(Bitmap bmp)
		{
			// Stvorimo VideoStream ako je potrebno
			if (aviFile == null)
			{
				try
				{
					aviFile = new AviWriter(fileName, compression, 25,
					                        bmp.Width, bmp.Height);
				}
				catch (Exception e)
				{
					aviFile = null;
				}
			}

			//
			if (aviFile != null)
			{
                try
                {
                    aviFile.WriteFrame(bmp);
                }
                catch
                {
                }
			}
		}
	}

	#endregion AVI

	/// <summary>
	/// 
	/// </summary>
	public class Visualize
	{
		/// <summary>
		/// 
		/// </summary>
		private Display currentDisplay;

		/// <summary>
		/// 
		/// </summary>
		private bool internalForceRefresh;

		//
		public uint PaintableComputerDisplaySize;
		public uint WirelessRadius;

		/// <summary>
		/// 
		/// </summary>
		private Rectangle bitmapSize;

        //
        private Rectangle displaySize;
        public Rectangle DisplaySize
        {
            get { return displaySize; }
        }

        private bool boldSensorsAndPortals = true;

        public bool BoldSensorsAndPortals
        {
            get { return boldSensorsAndPortals; }
            set { boldSensorsAndPortals = value; internalForceRefresh = true; }
        }

        /// <summary>
		/// 
		/// </summary>
		private Point originPoint;
		public Point OriginPoint
		{
			get { return originPoint; }
			set
			{
				if (currentDisplay != null)
					currentDisplay.ClearDisplay();
				originPoint = value;
			}
		}

        private double zoomFactor = 1.0;
        public double ZoomFactor
        {
            get { return zoomFactor; }
            set
            {
                if ((value >= 0.1) && (value <= 2.0))
                {
                    zoomFactor = value;

                    displaySize = new Rectangle(Convert.ToInt32(bitmapSize.X + originPoint.X * zoomFactor),
                                                Convert.ToInt32(bitmapSize.Y + originPoint.Y * zoomFactor),
                                                Convert.ToInt32(bitmapSize.Width * zoomFactor),
                                                Convert.ToInt32(bitmapSize.Height * zoomFactor));
                    if (currentDisplay != null)
                    {
                        currentDisplay.ClearDisplay();
                        internalForceRefresh = currentDisplay.ChangedZoom();
                    }
                }
            }
        }

        //
        public Point FixCoordinates(Point point)
        {
            return new Point(Convert.ToInt32((point.X / zoomFactor) + (originPoint.X / zoomFactor)),
                             Convert.ToInt32((point.Y / zoomFactor) + (originPoint.Y / zoomFactor)));
        }

		/// <summary>
		/// 
		/// </summary>
		private PConTheWall infoPC;

		private Point[] transparentRegion = null;
		private Point temporaryTransparentPoint = Point.Empty;

		/// <summary>
		/// Adds new point to polygon. If ReplaceLast is set, an extra, "moveing" point is
		/// defined to follow mouse movement.
		/// Poligon is freed when newPoint is null.
		/// </summary>
		/// <param name="newPoint"></param>
		/// <param name="ReplaceLast"></param>
		public int AddTransparentPoint(Point newPoint, bool ReplaceLast)
		{
			if (newPoint == Point.Empty)
			{
				transparentRegion = null;
				temporaryTransparentPoint = Point.Empty;
				internalForceRefresh = true;
				return 0;
			}

			if (transparentRegion == null)
			{
				if (!ReplaceLast)
				{
					transparentRegion = new Point[1];
					transparentRegion[0] = newPoint;

					internalForceRefresh = true;
					temporaryTransparentPoint = Point.Empty;
				}
			}
			else
			{
				if (transparentRegion.Length < 2)
				{
					if (!ReplaceLast)
					{
						Point[] tmpTransparentRegion;
						tmpTransparentRegion = new Point[transparentRegion.Length + 1];
						for (int i = 0; i < transparentRegion.Length; i++)
							tmpTransparentRegion[i] = transparentRegion[i];
						transparentRegion = tmpTransparentRegion;

						transparentRegion[transparentRegion.Length - 1] = newPoint;
						temporaryTransparentPoint = Point.Empty;
					}
				}
				else
				{
					if (!ReplaceLast)
					{
						Point[] tmpTransparentRegion;
						tmpTransparentRegion = new Point[transparentRegion.Length + 1];
						for (int i = 0; i < transparentRegion.Length; i++)
							tmpTransparentRegion[i] = transparentRegion[i];
						transparentRegion = tmpTransparentRegion;

						transparentRegion[transparentRegion.Length - 1] = newPoint;
						temporaryTransparentPoint = Point.Empty;
					}
					else
						temporaryTransparentPoint = newPoint;

					internalForceRefresh = true;
				}
			}
			//
			if (transparentRegion == null)
				return 0;
			else
				return transparentRegion.Length;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public Point[] GetTransparentPoints()
		{
			return transparentRegion;
		}

		/// <summary>
		/// Constructor
		/// </summary>
		public Visualize()
		{
			currentDisplay = null;
			internalForceRefresh = true;
			infoPC = null;
			transparentRegion = null;

			//Default vrijednosti
			PaintableComputerDisplaySize = 3;
			WirelessRadius = 50;

			//
			originPoint = new Point(0, 0);
		}

		/// <summary>
		/// Destructor
		/// </summary>
		~Visualize()
		{
			Terminate();
		}

		/// <summary>
		/// Performs cleanup
		/// </summary>
		public void Terminate()
		{
			if (currentDisplay != null)
			{
                DoPaint();
				currentDisplay.Cleanup();
				currentDisplay = null;
			}
		}

        //
        public void RefreshDisplaySize()
        {
            if (currentDisplay == null) return;
            currentDisplay.RefreshDisplay(bitmapSize);
        }

		/// <summary>
		/// 
		/// </summary>
		/// <param name="displayType"></param>
		/// <param name="Window"></param>
		public void SetDisplayWindow(Type displayType,
		                             IntPtr Window,
		                             Size WallSize)
		{
			if (currentDisplay != null)
				currentDisplay.Cleanup();

			currentDisplay = (Display) Activator.CreateInstance(displayType, new Object[] {Window});
			internalForceRefresh = true;

			bitmapSize = new Rectangle(0, 0, WallSize.Width, WallSize.Height);
            ZoomFactor = 1.0;
		}

		public void SetInfoPC(PConTheWall InfoPC)
		{
			if (infoPC != InfoPC)
			{
				infoPC = InfoPC;
				internalForceRefresh = true;
			}
		}
        public PConTheWall GetInfoPC()
        {
            return infoPC;
        }

		/// <summary>
		/// Draws Paintable computers on bitmap using selected Display
		/// </summary>
		/// <param name="computerList"></param>
		/// <param name="forceDraw"></param>
		public void DoDraw(IList<PConTheWall> computerList,
		                   List<Color> colors,
		                   bool forceDraw)
		{
			if (currentDisplay == null) return;

			forceDraw = forceDraw || internalForceRefresh;
			internalForceRefresh = false;

			currentDisplay.BeginUpdate(forceDraw);
			PConTheWall pc;

			for (int i = 0; i < computerList.Count; i++)
			{
				pc = computerList[i];


				Color pcColor;
                if (pc.IsAlive)
                {
                    pcColor = colors[i];
                }
                else
                    pcColor = Color.FromArgb(50, 50, 50);

				// Jo jedan trik, ako PC nije promjenio boju
				// nemoj troiti vrijeme na iscrtavanje
				if ((pc.LastColor != pcColor) || (forceDraw))
				{
					pc.LastColor = pcColor;
                    // Infiltrirani PC je drukije od ostalih
                    if (boldSensorsAndPortals)
                    {
                        if (pc.IsPortal)
                        {
                            currentDisplay.DrawPoint(pc.Location, PaintableComputerDisplaySize * 2, Color.Yellow);
                        }
                        else if (pc.IsSensor)
                        {
                            currentDisplay.DrawPoint(pc.Location, PaintableComputerDisplaySize * 2, Color.Red);
                        }
                    }
                    currentDisplay.DrawPoint(pc.Location, PaintableComputerDisplaySize, pcColor);
				}
			}

			// E, ovo je za info PC
			// crtamo radijus  susjedne PCje
			if (infoPC != null)
			{
				// Oznaeni PC
				Color pcColor = Color.Yellow;
				if ((infoPC.LastColor != pcColor) || (forceDraw))
				{
					infoPC.LastColor = pcColor;
					currentDisplay.DrawPoint(infoPC.Location, PaintableComputerDisplaySize, pcColor);
				}

				// Susjedi
				pcColor = Color.YellowGreen;
				PConTheWall pcNeighbours;
				for (int j = 0; j < infoPC.Neighbours.Count; j++)
				{
					pcNeighbours = infoPC.Neighbours[j];
					if ((pcNeighbours.LastColor != pcColor) || (forceDraw))
					{
						pcNeighbours.LastColor = pcColor;
						currentDisplay.DrawPoint(pcNeighbours.Location, PaintableComputerDisplaySize, pcColor);
					}
				}
				currentDisplay.DrawCircle(infoPC.Location, WirelessRadius, Color.Yellow);
			}

			if (transparentRegion != null)
			{
				Point[] tmpTransparentRegion;
				if (temporaryTransparentPoint != POINT.Empty)
				{
					tmpTransparentRegion = new Point[transparentRegion.Length + 1];
					for (int i = 0; i < transparentRegion.Length; i++)
						tmpTransparentRegion[i] = transparentRegion[i];
					tmpTransparentRegion[tmpTransparentRegion.Length - 1] = temporaryTransparentPoint;
				}
				else
					tmpTransparentRegion = transparentRegion;
				currentDisplay.DrawTransparentShape(tmpTransparentRegion, Color.Red, 90);
			}

			currentDisplay.EndUpdate();
		}

		/// <summary>
		/// Actually paints bitmap on selected windows handle
		/// </summary>
		public void DoPaint()
		{
			if (currentDisplay == null) return;
            currentDisplay.DisplayOnScreen(displaySize);
		}

		/// <summary>
		/// Returns only bitmap (used mostily in ASP.NET project)
		/// </summary>
		/// <returns></returns>
		public Bitmap ReturnBitmap()
		{
			if (currentDisplay != null)
				return (Bitmap) currentDisplay.hBmp.Clone();
			else
				return null;
		}
	}

	// Tko eli raditi prikaz rezultata u DirectXu ili OpenGLu
	// slobodno neka naslijedi ovu klasu neka radi cuda :0)

	#region Abstract Display

	public abstract class Display
	{
		// Forsranje brisanja ekrana
		protected bool fForceClear = false;

		public void ForceClear()
		{
			fForceClear = true;
		}

		// Privatne varijable
		protected IntPtr hWindow;
		public Bitmap hBmp;
		protected Point[] tmpPoint;

        protected Rectangle rBitmap;
		protected Graphics hDisplay;
		protected Graphics hGraphics;


		// Konstruktor
		// Prima samo handle prozora gdje se iscrtava
		public Display(IntPtr Window)
		{
			hWindow = Window;
		}

		// Nakon sto se display stvori od prozora
		// fixiran je na njegovu tadasnju velicinu...
		// Ovo treba pozvati nakon resajziranja prozora
		public abstract void RefreshDisplay(Rectangle bitmapSize);
		public abstract void Cleanup();
		// Virtualne funkcije koje treba ostvariti
		public abstract void BeginUpdate(bool ForceNew);
		public abstract void EndUpdate();
        public abstract bool ChangedZoom();
		public abstract void Clear();
		public abstract void ClearDisplay();
        public abstract void DisplayOnScreen(Rectangle DisplaySize);
		public abstract void DrawLine(Point FromPoint, Point ToPoint);
		public abstract void DrawPoint(Point Centar, uint Radius, Color PointColor);
		public abstract void DrawCircle(Point Centar, uint Radius, Color PointColor);
		public abstract void DrawTransparentShape(Point[] Points, Color PointColor, int Transparency);
	}

	#endregion Abstract Display

	// Prva implementacija koristei GDI+ funkcije    

	#region GDI+

	public class DisplayGDIPlus : Display
	{
		//HACK
		private Color   tmpPointColor;
		private int     tmpTransparency;
		private bool    doUpdateTransparentRegion;

		// Konstruktor
		public DisplayGDIPlus(IntPtr Window)
			: base(Window)
		{
			if (Window != IntPtr.Zero)
				hDisplay = Graphics.FromHwnd(hWindow);
			else
				hDisplay = null;
		}

		public override void Cleanup()
		{
			if (hBmp != null)
			{
				hGraphics.Dispose();
				hBmp.Dispose();
			}
			if (hDisplay != null)
				hDisplay.Dispose();
		}

		~DisplayGDIPlus()
		{
			Cleanup();
		}

		public override void RefreshDisplay(Rectangle bitmapSize)
		{
			if (hDisplay != null)
				hDisplay.Dispose();

            if ((hWindow != IntPtr.Zero))
            {
                if (rBitmap != bitmapSize)
                {
                    rBitmap = bitmapSize;

                    if (hBmp != null)
                    {
                        hGraphics.Dispose();
                        hBmp.Dispose();
                    }

                    //PixelFormat.Format24bppRgb
                    hBmp = new Bitmap(rBitmap.Width, rBitmap.Height, PixelFormat.Format32bppPArgb);
                    hGraphics = Graphics.FromImage(hBmp);
                }
                hDisplay = Graphics.FromHwnd(hWindow);
            }
            else
            {
                hDisplay = null;
            }
		}


		// Funkcije iscrtavanja
		public override void DrawLine(Point FromPoint, Point ToPoint)
		{
			Pen p = new Pen(Color.Red);
			try
			{
				hGraphics.DrawLine(new Pen(Color.Red), FromPoint, ToPoint);
			}
			finally
			{
				p.Dispose();
			}
		}

		public override void DrawPoint(Point Centar, uint Radius, Color PointColor)
		{
			Brush b = new SolidBrush(PointColor);
			try
			{
				hGraphics.FillEllipse(b, Centar.X - Radius,
				                      Centar.Y - Radius,
				                      2*Radius, 2*Radius);
			}
			finally
			{
				b.Dispose();
			}
		}

		public override void DrawCircle(Point Centar, uint Radius, Color PointColor)
		{
			Pen p = new Pen(PointColor);
			try
			{
				hGraphics.DrawEllipse(p,
				                      Centar.X - Radius,
				                      Centar.Y - Radius,
				                      Radius<<1, Radius<<1);
			}
			finally
			{
				p.Dispose();
			}
		}

		public override void Clear()
		{
			hGraphics.Clear(Color.Black);
		}

		public override void ClearDisplay()
		{
			hDisplay.Clear(Color.Black);
		}

		public override void BeginUpdate(bool ForceNew)
		{
			if (ForceNew)
			{
				Clear();
			}
			if (fForceClear)
			{
				if (hDisplay != null)
					hDisplay.Clear(Color.Black);
				fForceClear = false;
			}
		}

		public override void EndUpdate()
		{
			// Kod iz DrawTransparentShape
			// Points -> tmpPoint
			if ((tmpPoint != null) && (doUpdateTransparentRegion))
			{
				Brush b = new SolidBrush(Color.FromArgb(tmpTransparency, tmpPointColor.R, tmpPointColor.G, tmpPointColor.B));
				try
				{
					hGraphics.FillPolygon(b, tmpPoint);
				}
				finally
				{
					b.Dispose();
				}
				doUpdateTransparentRegion = false;
			}
			//
			//if (hDisplay != null)
			//    hDisplay.DrawImage(hBmp, rect);
		}

		public override void DisplayOnScreen(Rectangle DisplaySize)
		{
            if (hDisplay != null)
                hDisplay.DrawImage(hBmp, DisplaySize,
                                         rBitmap,
                                         GraphicsUnit.Pixel);
		}

		public override void DrawTransparentShape(Point[] Points, Color PointColor, int Transparency)
		{
			doUpdateTransparentRegion = (tmpPoint != Points);
			tmpPoint = Points;
			tmpPointColor = PointColor;
			tmpTransparency = Transparency;

			/* HACK!
             * GDI+ nema XOR funkciju pa se ne moe izbrisati neto
             * to je ve na slici. Zato u ovoj funkciji samo pamtimo
             * koju regiju treba iscrtati (prva linija), a kod koji
             * to zapravo crta prebacijemo u EndUpdate() fju.
             *
			if (Points == null) return;

			Brush b = new SolidBrush( Color.FromArgb(Transparency, PointColor.R, PointColor.G, PointColor.B));
            try
            {
            	hGraphics.FillPolygon(b, Points);
            }
            finally
            {
                b.Dispose();
            }
             */
		}
        public override bool ChangedZoom()
        {
            return false;
        }
    }

	#endregion GDI+

	// Druga, koristei standardne GDI32 funkcije...
	// Nevjerojatno brza preko Remote Desktop veze
	// otprilike duplo bra od GDI+

	#region GDI
	public class DisplayGDI : Display
	{
		private Graphics hDisplay;
        private Rectangle rDisplay;

		private Graphics hGraphics;
		private Rectangle bmpSize;


		private IntPtr bmpDC;
		private Color LastColor;
		private HBRUSH brush;
		private HPEN pen;

		// Konstruktor
		public DisplayGDI(IntPtr Window)
			: base(Window)
		{
			if (Window != IntPtr.Zero)
				hDisplay = Graphics.FromHwnd(hWindow);
			else
				hDisplay = null;
		}

		public override void Cleanup()
		{
			if (hBmp != null)
			{
				try
				{
					hGraphics.ReleaseHdc(bmpDC);
					hGraphics.Dispose();
					hBmp.Dispose();
				}
				catch
				{
				}
			}
			if (hDisplay != null)
				hDisplay.Dispose();
		}

		~DisplayGDI()
		{
			Cleanup();
		}

        public override bool ChangedZoom()
        {
            return true;
        }
        public override void RefreshDisplay(Rectangle bitmapSize)
		{
			if (hDisplay != null)
				hDisplay.Dispose();

			if (hWindow != IntPtr.Zero)
            {
                bmpSize = bitmapSize;

                if (hBmp != null)
                {
                    hGraphics.ReleaseHdc(bmpDC);
                    hGraphics.Dispose();
                    hBmp.Dispose();
                }

                //PixelFormat.Format24bppRgb 
                hBmp = new Bitmap(bmpSize.Width, bmpSize.Height, PixelFormat.Format24bppRgb);
                hGraphics = Graphics.FromImage(hBmp);

                bmpDC = new IntPtr();
                bmpDC = hGraphics.GetHdc();

                ChechBrushAndPen(Color.Black, false);
                ClearDisplay();
            }
			else
            {
				hDisplay = null;
            }
		}

		private void ChechBrushAndPen(Color c, bool Hatched)
		{
			if (LastColor != c)
			{
				if (brush != (HBRUSH) 0)
				{
					GDI.DeleteObject(brush);
					GDI.DeleteObject(pen);
				}
				LastColor = c;

				if (!Hatched)
					brush = (HBRUSH) GDI.CreateSolidBrush(LastColor.B << 16 | LastColor.G << 8 | LastColor.R);
				else
					brush = (HBRUSH) GDI.CreateHatchBrush(GDI.HS_SOLID, LastColor.B << 16 | LastColor.G << 8 | LastColor.R);

				pen = (HPEN) GDI.CreatePen(GDI.PS_SOLID, 2, LastColor.B << 16 | LastColor.G << 8 | LastColor.R);

				GDI.SelectObject(bmpDC, brush);
				GDI.SelectObject(bmpDC, pen);
			}
		}


		// Funkcije iscrtavanja
		public override void DrawLine(Point FromPoint, Point ToPoint)
		{
		}

		public override void DrawPoint(Point Centar, uint Radius, Color PointColor)
		{
			ChechBrushAndPen(PointColor, false);

			GDI.Ellipse(bmpDC, (int) (Centar.X - Radius), (int) (Centar.Y - Radius),
			            (int) (Centar.X + Radius), (int) (Centar.Y + Radius));
		}

		public override void DrawCircle(Point Centar, uint Radius, Color PointColor)
		{
			ChechBrushAndPen(PointColor, false);

			int x1 = (int) (Centar.X - Radius);
			int y1 = (int) (Centar.Y - Radius);
			int x2 = (int) (Centar.X + Radius);
			int y2 = (int) (Centar.Y + Radius);

			GDI.Arc(bmpDC, x1, y1, x2, y2,
			        x1, y1, x1, y1);
		}

		public override void Clear()
		{
            LastColor = Color.Wheat;
            ChechBrushAndPen(Color.Black, false);
			GDI.Rectangle(bmpDC, 0, 0, bmpSize.Width, bmpSize.Width);
		}

		public override void ClearDisplay()
		{
            if (hDisplay != null)
                hDisplay.Dispose();

            hDisplay = Graphics.FromHwnd(hWindow);
            rDisplay = new Rectangle(Convert.ToInt32(hDisplay.VisibleClipBounds.X),
                                 Convert.ToInt32(hDisplay.VisibleClipBounds.Y),
                                 Convert.ToInt32(hDisplay.VisibleClipBounds.Width),
                                 Convert.ToInt32(hDisplay.VisibleClipBounds.Height));
            hDisplay.Clear(Color.Black);
		}

		public override void BeginUpdate(bool ForceNew)
		{
			if (ForceNew)
			{
				Clear();
			}
		}

		public override void EndUpdate()
		{
			//
		}

		public override void DisplayOnScreen(Rectangle source)
		{
			if (hDisplay != null)
			{
				IntPtr dDC = new IntPtr();
				dDC = hDisplay.GetHdc();

                if (bmpSize == source)
                    GDI.BitBlt(dDC, 0, 0, rDisplay.Width, rDisplay.Height,
                               bmpDC, source.Left, source.Top,
                               GDI.SRCCOPY);
                else
                {
                    GDI.StretchBlt(dDC, 0, 0, rDisplay.Width, rDisplay.Height,
                                   bmpDC, 0, 0, bmpSize.Width, bmpSize.Height,
                                   GDI.SRCCOPY);
                }

                hDisplay.ReleaseHdc(dDC);
			}
		}

		public override void DrawTransparentShape(Point[] Points, Color PointColor, int Transparency)
		{
			Color c = LastColor;
			ChechBrushAndPen(PointColor, true);

			GDI.SetROP2(bmpDC, GDI.R2_XORPEN);
			if (tmpPoint != null)
			{
				GDI.Polygon(bmpDC, ref tmpPoint[0], tmpPoint.Length);
			}
			if (Points == null)
				tmpPoint = null;
			else
			{
				tmpPoint = Points;

				//new Point[Points.Length];
				//Points.CopyTo(tmpPoint, 0);
			}
			GDI.SetROP2(bmpDC, GDI.R2_COPYPEN);
			if (Points != null)
				GDI.Polygon(bmpDC, ref Points[0], Points.Length);
			ChechBrushAndPen(c, false);
		}
	}
 	#endregion GDI
}